from typing import *  # necessary for eval()

from pedantic.type_checking_logic.check_types import get_type_arguments
from pedantic.exceptions import PedanticDocstringException
from pedantic.models.decorated_function import DecoratedFunction


def _check_docstring(decorated_func: DecoratedFunction) -> None:
    doc = decorated_func.docstring
    err = decorated_func.err
    context = {}

    _assert_docstring_is_complete(func=decorated_func)

    for annotation in decorated_func.annotations:
        expected_type = decorated_func.annotations[annotation]
        _update_context(context=context, type_=expected_type)

        if annotation == 'return' and decorated_func.annotations[annotation] is not None:
            if len(doc.returns.args) != 2:
                raise PedanticDocstringException(f'{err} Parsing Error. Only Google style docstrings are supported.')

            actual_return_type = _parse_documented_type(type_=doc.returns.args[1], context=context, err=err)

            if actual_return_type != expected_type:
                raise PedanticDocstringException(
                    f'{err} Documented type is incorrect: Annotation: {expected_type} Documented: {actual_return_type}')
        elif annotation != 'return':
            docstring_param = list(filter(lambda p, a=annotation: p.arg_name == a, doc.params))[0]
            actual_param_type = _parse_documented_type(type_=docstring_param.type_name, context=context, err=err)

            if expected_type != actual_param_type:
                raise PedanticDocstringException(
                    f'{err} Documented type of parameter {annotation} is incorrect. Expected {expected_type} '
                    f'but documented is {actual_param_type}.')


def _assert_docstring_is_complete(func: DecoratedFunction) -> None:
    """
        >>> def f1(): pass
        >>> _assert_docstring_is_complete(DecoratedFunction(f1))
        Traceback (most recent call last):
        ...
        pedantic.exceptions.PedanticDocstringException: In function f1:
         The function should have a docstring!
        >>> def f2():
        ...     ''' This is a docstring. '''
        ...     pass
        >>> _assert_docstring_is_complete(DecoratedFunction(f2))
        >>> def f3() -> None:
        ...     ''' This is a docstring. '''
        ...     pass
        >>> _assert_docstring_is_complete(DecoratedFunction(f3))
        >>> def f4() -> int:
        ...     ''' This is a docstring. '''
        ...     return 42
        >>> _assert_docstring_is_complete(DecoratedFunction(f4))
        Traceback (most recent call last):
        ...
        pedantic.exceptions.PedanticDocstringException: In function f4:
         The return type <class 'int'> is not documented.
        >>> def f5() -> int:
        ...     ''' This is a docstring.
        ...     Returns:
        ...         int: a magic number
        ...     '''
        ...     return 42
        >>> _assert_docstring_is_complete(DecoratedFunction(f5))
    """
    if func.raw_doc is None or func.raw_doc == '':
        raise PedanticDocstringException(f'{func.err} The function should have a docstring!')

    num_documented_args = len(func.docstring.params)
    num_taken_args = len([a for a in func.annotations if a != 'return'])

    if num_documented_args != num_taken_args:
        raise PedanticDocstringException(
            f'{func.err} There are {num_documented_args} argument(s) documented, '
            f'but {num_taken_args} are actually taken.')

    if func.docstring.returns is None and ('return' in func.annotations and func.annotations['return'] is not None):
        raise PedanticDocstringException(f'{func.err} The return type {func.annotations["return"]} is not documented.')

    if func.docstring.returns is not None and ('return' not in func.annotations or func.annotations['return'] is None):
        raise PedanticDocstringException(
            f'{func.err} The return type {func.docstring.returns.type_name} '
            f'is documented but the function does not return anything.')


def _parse_documented_type(type_: str, context: Dict[str, Any], err: str) -> Any:
    """
    >>> import sys
    >>> _parse_documented_type(type_='List[str]', context={}, err='')
    typing.List[str]
    >>> _parse_documented_type(type_='float', context={}, err='')
    <class 'float'>
    >>> _parse_documented_type(type_='List[List[bool]]', context={}, err='')
    typing.List[typing.List[bool]]
    >>> _parse_documented_type(type_='Union[int, float, str]', context={}, err='')
    typing.Union[int, float, str]
    >>> _parse_documented_type(type_='Callable[[int, bool, str], float]', context={}, err='')
    typing.Callable[[int, bool, str], float]
    >>> _parse_documented_type(type_='Optional[List[Dict[str, float]]]', context={}, err='')
    typing.Optional[typing.List[typing.Dict[str, float]]]
    >>> _parse_documented_type(type_='Optional[List[Dict[str, float]]]', context={}, err='')
    typing.Optional[typing.List[typing.Dict[str, float]]]
    >>> _parse_documented_type(type_='Union[List[Dict[str, float]], None]', context={}, err='')
    typing.Optional[typing.List[typing.Dict[str, float]]]
    >>> _parse_documented_type(type_='Union[List[Dict[str, float]], None]', context={}, err='')
    typing.Optional[typing.List[typing.Dict[str, float]]]
    >>> _parse_documented_type(type_='MyClass', context={}, err='')
    Traceback (most recent call last):
    ...
    pedantic.exceptions.PedanticDocstringException: Documented type "MyClass" was not found.
    >>> class MyClass: pass
    >>> _parse_documented_type(type_='MyClass', context={'MyClass': MyClass}, err='')
    <class 'pedantic.type_checking_logic.check_docstring.MyClass'>
    >>> _parse_documented_type(type_='MyClas', context={'MyClass': MyClass}, err='')
    Traceback (most recent call last):
    ...
    pedantic.exceptions.PedanticDocstringException: Documented type "MyClas" was not found. Maybe you meant "MyClass"
    >>> class MyClub: pass
    >>> _parse_documented_type(type_='MyClas', context={'MyClass': MyClass, 'MyClub': MyClub}, err='')
    Traceback (most recent call last):
    ...
    pedantic.exceptions.PedanticDocstringException: Documented type "MyClas" was not found. Maybe you meant one of the following: ['MyClass', 'MyClub']
    """

    if 'typing.' in type_:
        raise PedanticDocstringException(
            f'{err}Do not use "typing." in docstring. Please replace "{type_}" with '
            f'"{type_.replace("typing.", "")}" in the docstring')

    try:
        return eval(type_, globals(), context)
    except NameError:
        possible_meant_types = [t for t in context.keys() if isinstance(t, str)]
        if len(possible_meant_types) > 1:
            msg = f' Maybe you meant one of the following: {possible_meant_types}'
        elif len(possible_meant_types) == 1:
            msg = f' Maybe you meant "{possible_meant_types[0]}"'
        else:
            msg = ''
        raise PedanticDocstringException(f'{err}Documented type "{type_}" was not found.{msg}')


def _update_context(context: Dict[str, Any], type_: Any) -> Dict[str, Any]:
    """
        >>> from typing import List, Union, Optional, Callable
        >>> _update_context(type_=None, context={})
        {}
        >>> _update_context(type_=None, context={'a': 1, 'b': 2})
        {'a': 1, 'b': 2}
        >>> _update_context(type_=float, context={})
        {'float': <class 'float'>}
        >>> _update_context(type_=List[str], context={})
        {'str': <class 'str'>}
        >>> _update_context(type_=List[List[bool]], context={})
        {'bool': <class 'bool'>}
        >>> _update_context(type_=Union[int, float, str], context={})
        {'int': <class 'int'>, 'float': <class 'float'>, 'str': <class 'str'>}
        >>> _update_context(type_=Callable[[int, bool, str], float], context={})
        {'int': <class 'int'>, 'bool': <class 'bool'>, 'str': <class 'str'>, 'float': <class 'float'>}
        >>> _update_context(type_=Optional[List[Dict[str, float]]], context={})
        {'str': <class 'str'>, 'float': <class 'float'>, 'NoneType': <class 'NoneType'>}
        >>> _update_context(type_=Union[List[Dict[str, float]], None], context={})
        {'str': <class 'str'>, 'float': <class 'float'>, 'NoneType': <class 'NoneType'>}
        >>> _update_context(type_='MyClass', context={})
        {'MyClass': 'MyClass'}
        >>>
    """

    if isinstance(type_, str):
        context[type_] = type_
    elif str(type_).startswith('typing'):
        type_arguments = get_type_arguments(cls=type_)

        for type_argument in type_arguments:
            if isinstance(type_argument, list):  # Callable has a List of Types as first argument
                for type_arg in type_argument:
                    _update_context(context=context, type_=type_arg)
            _update_context(context=context, type_=type_argument)
    elif hasattr(type_, '__name__'):
        context[type_.__name__] = type_
    return context



from unittest import TestCase

from pedantic.decorators.fn_deco_validate.exceptions import ParameterException
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate
from pedantic.decorators.fn_deco_validate.parameters import Parameter
from pedantic.decorators.fn_deco_validate.validators import MinLength


class TestValidatorMinLength(TestCase):
    def test_validator_min_length(self) -> None:
        @validate(Parameter(name='x', validators=[MinLength(3)]))
        def foo(x):
            return x

        self.assertEqual('hi!', foo('hi!'))
        self.assertEqual('hello', foo('hello'))
        self.assertEqual([1, 2, 3], foo([1, 2, 3]))

        with self.assertRaises(expected_exception=ParameterException):
            foo('hi')

        with self.assertRaises(expected_exception=ParameterException):
            foo([1, 2])

        with self.assertRaises(expected_exception=ParameterException):
            foo(42)



import unittest

from pedantic import DecoratorType, create_decorator, WithDecoratedMethods


class Decorators(DecoratorType):
    FOO = '_foo'
    BAR = '_bar'


foo = create_decorator(decorator_type=Decorators.FOO)
bar = create_decorator(decorator_type=Decorators.BAR)


class TestWithDecoratedMethods(unittest.TestCase):
    def test_with_decorated_methods_sync(self):
        class MyClass(WithDecoratedMethods[Decorators]):
            @foo(42)
            def m1(self) -> None:
                print('bar')

            @foo(value=43)
            def m2(self) -> None:
                print('bar')

            @bar(value=44)
            def m3(self) -> None:
                print('bar')

        instance = MyClass()
        expected = {
            Decorators.FOO: {
                instance.m1: 42,
                instance.m2: 43,
            },
            Decorators.BAR: {
                instance.m3: 44,
            }
        }
        assert instance.get_decorated_functions() == expected

    def test_with_decorated_methods_async(self):
        class MyClass(WithDecoratedMethods[Decorators]):
            @foo(42)
            async def m1(self) -> None:
                print('bar')

            @foo(value=43)
            async def m2(self) -> None:
                print('bar')

            @bar(value=44)
            async def m3(self) -> None:
                print('bar')

        instance = MyClass()
        expected = {
            Decorators.FOO: {
                instance.m1: 42,
                instance.m2: 43,
            },
            Decorators.BAR: {
                instance.m3: 44,
            }
        }
        assert instance.get_decorated_functions() == expected



from functools import wraps
from typing import Callable

from pedantic.constants import F, ReturnType


class Rename:
    def __init__(self, from_: str, to: str) -> None:
        self.from_ = from_
        self.to = to


def rename_kwargs(*params: Rename) -> Callable[[F], F]:
    """
        Renames the keyword arguments based on the given "Rename" rules when the decorated function is called.
        You can also use this to define aliases for keyword arguments.

        Example:

        >>> @rename_kwargs(
        ...     Rename(from_='firstname', to='a'),
        ...     Rename(from_='lastname', to='b'),
        ... )
        ... def my_function(a, b):
        ...     return a + ' ' + b
        >>> my_function(a='egon', b='olsen')  # the normal way
        'egon olsen'
        >>> my_function(firstname='egon', lastname='olsen')  # using new defined keyword arguments
        'egon olsen'
    """

    param_dict = {p.from_: p.to for p in params}

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args, **kwargs) -> ReturnType:
            result_kwargs = {}

            for k, v in kwargs.items():
                if k in param_dict:
                    result_kwargs[param_dict[k]] = kwargs[k]
                else:
                    result_kwargs[k] = kwargs[k]

            return func(*args, **result_kwargs)
        return wrapper
    return decorator


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from os import path

from setuptools import find_packages, setup


def get_content_from_readme(file_name: str = 'README.md') -> str:
    this_directory = path.abspath(path.dirname(__file__))

    with open(path.join(this_directory, file_name), encoding='utf-8') as file:
        return file.read()


url = "https://github.com/LostInDarkMath/pedantic-python-decorators"
author = "Willi Sontopski"

setup(
    name="pedantic",
    version="2.1.4",
    python_requires='>=3.11.0',
    packages=find_packages(),
    install_requires=[],
    author=author,
    author_email="willi_sontopski@arcor.de",
    license="Apache-2.0 License",
    maintainer=author,
    description="Some useful Python decorators for cleaner software development.",
    long_description=get_content_from_readme(),
    long_description_content_type='text/markdown',
    keywords="decorators tools helpers type-checking pedantic type annotations",
    url=url,
    project_urls={
        "Bug Tracker": f'{url}/issues',
        "Documentation": 'https://lostindarkmath.github.io/pedantic-python-decorators/pedantic/',
        "Source Code": url,
    },
    include_package_data=False,
    zip_safe=True,
)



from typing import Union

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class Min(Validator):
    def __init__(self, value: Union[int, float], include_boundary: bool = True) -> None:
        """
            >>> Min(7, True).validate(7)
            True
            >>> Min(7, False).validate(7)
            False
            >>> Min(7, False).validate(7.001)
            True
        """
        self._value = value
        self._include_boundary = include_boundary

    @overrides(Validator)
    def validate(self, value: Union[int, float]) -> Union[int, float]:
        if value < self._value and self._include_boundary:
            self.raise_exception(msg=f'smaller then allowed: {value} is not >= {self._value}', value=value)
        elif value <= self._value and not self._include_boundary:
            self.raise_exception(msg=f'smaller then allowed: {value} is not > {self._value}', value=value)

        return value



from abc import ABC, abstractmethod
from typing import Any, Dict, Type

from flask import request

from pedantic.decorators.fn_deco_validate.exceptions import InvalidHeader, ParameterException, ValidatorException
from pedantic.decorators.fn_deco_overrides import overrides
from pedantic.decorators.fn_deco_validate.parameters import ExternalParameter, Parameter
from pedantic.decorators.fn_deco_validate.parameters.deserializable import Deserializable


class FlaskParameter(ExternalParameter, ABC):
    @abstractmethod
    def get_dict(self) -> Dict[str, Any]:
        """ Returns the actual values as a dictionary. """

    @overrides(ExternalParameter)
    def has_value(self) -> bool:
        dict_ = self.get_dict()
        return dict_ is not None and self.name in dict_

    @overrides(ExternalParameter)
    def load_value(self) -> Any:
        dict_ = self.get_dict()
        return dict_[self.name]


class FlaskJsonParameter(FlaskParameter):
    @overrides(FlaskParameter)
    def get_dict(self) -> Dict:
        if not request.is_json:
            return {}

        return request.json


class FlaskFormParameter(FlaskParameter):
    @overrides(FlaskParameter)
    def get_dict(self) -> Dict:
        return request.form


class FlaskPathParameter(Parameter):
    """
    This is a special case because Flask passes path parameter as kwargs to validate().
    Therefore, this doesn't need to be an ExternalParameter.
    """


class FlaskGetParameter(FlaskParameter):
    @overrides(FlaskParameter)
    def get_dict(self) -> Dict:
        return request.args

    @overrides(ExternalParameter)
    def load_value(self) -> Any:
        value = request.args.getlist(self.name)

        if self.value_type == list:
            return value

        return value[0]


class FlaskHeaderParameter(FlaskParameter):
    exception_type = InvalidHeader

    @overrides(FlaskParameter)
    def get_dict(self) -> Dict:
        return request.headers


class GenericFlaskDeserializer(ExternalParameter):
    """
        A JSON deserializer for classes which implements the [Deserializable] interface.

        Further reading: https://github.com/LostInDarkMath/pedantic-python-decorators/issues/55
    """

    def __init__(self, cls: Type[Deserializable], catch_exception: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self._cls = cls
        self._catch_exceptions = catch_exception

    @overrides(ExternalParameter)
    def has_value(self) -> bool:
        return request.is_json

    @overrides(ExternalParameter)
    def load_value(self) -> Any:
        try:
            return self._cls.from_json(request.json)
        except ValidatorException as ex:
            raise ParameterException.from_validator_exception(exception=ex, parameter_name='')
        except Exception as ex:
            if self._catch_exceptions:
                self.raise_exception(msg=str(ex))

            raise ex



from .abstract_parameter import Parameter
from .abstract_external_parameter import ExternalParameter
from .deserializable import Deserializable
from .environment_variable_parameter import EnvironmentVariableParameter

try:
    from .flask_parameters import FlaskJsonParameter, FlaskFormParameter, FlaskParameter, FlaskGetParameter, \
        FlaskPathParameter, FlaskHeaderParameter, GenericFlaskDeserializer
except ImportError:
    pass



from typing import Type, Any, Union

from pedantic.decorators.fn_deco_validate.exceptions import ConversionError

T = Union[bool, int, float, str, dict, list]


def convert_value(value: Any, target_type: Type[T]) -> T:
    if isinstance(value, target_type):
        return value

    value = str(value).strip().lower()

    if target_type == bool:
        if value in ['true', '1']:
            return True
        elif value in ['false', '0']:
            return False

        raise ConversionError(f'Value {value} cannot be converted to bool.')

    try:
        if target_type == list:
            return [item.strip() for item in value.split(',')]
        elif target_type == dict:
            value = {item.split(':')[0].strip(): item.partition(':')[-1].strip() for item in value.split(',')}

        return target_type(value)
    except ValueError:
        raise ConversionError(f'Value {value} cannot be converted to {target_type}.')



from copy import deepcopy
from dataclasses import dataclass, fields, replace
from typing import Type, TypeVar, Any, Union, Callable, Dict

from pedantic.get_context import get_context
from pedantic.type_checking_logic.check_types import assert_value_matches_type

T = TypeVar('T')


def frozen_type_safe_dataclass(cls: Type[T]) -> Type[T]:
    """ Shortcut for @frozen_dataclass(type_safe=True) """

    return frozen_dataclass(type_safe=True)(cls)


def frozen_dataclass(
        cls: Type[T] = None,
        type_safe: bool = False,
        order: bool = False,
        kw_only: bool = True,
        slots: bool = False,
) -> Union[Type[T], Callable[[Type[T]], Type[T]]]:
    """
        Makes the decorated class immutable and a dataclass by adding the [@dataclass(frozen=True)]
        decorator. Also adds useful copy_with() and validate_types() instance methods to this class (see below).

        If [type_safe] is True, a type check is performed for each field after the __post_init__ method was called
        which itself s directly called after the __init__ constructor.
        Note this have a negative impact on the performance. It's recommend to use this for debugging and testing only.

        In a nutshell, the followings methods will be added to the decorated class automatically:
        - __init__() gives you a simple constructor like "Foo(a=6, b='hi', c=True)"
        - __eq__() lets you compare objects easily with "a == b"
        - __hash__() is also needed for instance comparison
        - __repr__() gives you a nice output when you call "print(foo)"
        - copy_with() allows you to quickly create new similar frozen instances. Use this instead of setters.
        - deep_copy_with() allows you to create deep copies and modify them.
        - validate_types() allows you to validate the types of the dataclass.
                           This is called automatically when [type_safe] is True.

        If the [order] parameter is True (default is False), the following comparison methods
        will be added additionally:
        - __lt__() lets you compare instance like "a < b"
        - __le__() lets you compare instance like "a <= b"
        - __gt__() lets you compare instance like "a > b"
        - __ge__() lets you compare instance like "a >= b"

        These compare the class as if it were a tuple of its fields, in order.
        Both instances in the comparison must be of the identical type.

        The parameters slots and kw_only are only applied if the Python version is greater or equal to 3.10.

        Example:

        >>> @frozen_dataclass
        ... class Foo:
        ...     a: int
        ...     b: str
        ...     c: bool
        >>> foo = Foo(a=6, b='hi', c=True)
        >>> print(foo)
        Foo(a=6, b='hi', c=True)
        >>> print(foo.copy_with())
        Foo(a=6, b='hi', c=True)
        >>> print(foo.copy_with(a=42))
        Foo(a=42, b='hi', c=True)
        >>> print(foo.copy_with(b='Hello'))
        Foo(a=6, b='Hello', c=True)
        >>> print(foo.copy_with(c=False))
        Foo(a=6, b='hi', c=False)
        >>> print(foo.copy_with(a=676676, b='new', c=False))
        Foo(a=676676, b='new', c=False)
    """

    def decorator(cls_: Type[T]) -> Type[T]:
        args = {'frozen': True, 'order': order, 'kw_only': kw_only, 'slots': slots}

        if type_safe:
            old_post_init = getattr(cls_, '__post_init__', lambda _: None)

            def new_post_init(self) -> None:
                old_post_init(self)
                context = get_context(depth=3, increase_depth_if_name_matches=[
                    copy_with.__name__,
                    deep_copy_with.__name__,
                ])
                self.validate_types(_context=context)

            setattr(cls_, '__post_init__', new_post_init)  # must be done before applying dataclass()

        new_class = dataclass(**args)(cls_)  # slots = True will create a new class!

        def copy_with(self, **kwargs: Any) -> T:
            """
                Creates a new immutable instance that by copying all fields of this instance replaced by the new values.
                Keep in mind that this is a shallow copy!
            """

            return replace(self, **kwargs)

        def deep_copy_with(self, **kwargs: Any) -> T:
            """
                Creates a new immutable instance that by deep copying all fields of
                this instance replaced by the new values.
            """

            current_values = {field.name: deepcopy(getattr(self, field.name)) for field in fields(self)}
            return new_class(**{**current_values, **kwargs})

        def validate_types(self, *, _context: Dict[str, Type] = None) -> None:
            """
                Checks that all instance variable have the correct type.
                Raises a [PedanticTypeCheckException] if at least one type is incorrect.
            """

            props = fields(new_class)

            if _context is None:
                # method was called by user
                _context = get_context(depth=2)

            _context = {**_context, **self.__init__.__globals__, self.__class__.__name__: self.__class__}

            for field in props:
                assert_value_matches_type(
                    value=getattr(self, field.name),
                    type_=field.type,
                    err=f'In dataclass "{cls_.__name__}" in field "{field.name}": ',
                    type_vars={},
                    context=_context,
                )

        methods_to_add = [copy_with, deep_copy_with, validate_types]

        for method in methods_to_add:
            setattr(new_class, method.__name__, method)

        return new_class

    if cls is None:
        return decorator

    return decorator(cls_=cls)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



import sys
import unittest
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Optional, Callable, Union, Dict, List

from pedantic.env_var_logic import disable_pedantic, enable_pedantic
from pedantic import overrides
from pedantic.decorators.class_decorators import pedantic_class
from pedantic.exceptions import PedanticOverrideException, PedanticTypeCheckException, \
    PedanticCallWithArgsException


class TestPedanticClass(unittest.TestCase):
    def tearDown(self) -> None:
        enable_pedantic()

    def test_constructor(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

        MyClass(a=42)

    def test_constructor_with_list(self):
        class Foo(IntEnum):
            A = 1
            B = 2

        @pedantic_class
        class MyClass:
            def __init__(self, b: int, a: List[Foo]) -> None:
                self.a = a
                self.b = b

        MyClass(b=42, a=[Foo.A, Foo.B])

    def test_constructor_param_without_type_hint(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a) -> None:
                self.a = a

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass(a=42)

    def test_constructor_without_return_type(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int):
                self.a = a

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass(a=42)

    def test_constructor_wrong_return_type(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> int:
                self.a = a

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass(a=42)

    def test_constructor_must_be_called_with_kwargs(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            MyClass(42)

    def test_multiple_methods(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b: int) -> int:
                return self.a - b

            def print(self, s: str) -> None:
                print(f'{self.a} and {s}')

        m = MyClass(a=5)
        m.calc(b=42)
        m.print(s='Hi')
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.calc(45)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.calc(b=45.0)
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.print('Hi')

    def test_multiple_methods_with_missing_and_wrong_type_hints(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b: int) -> float:
                return self.a - b

            def dream(self, b) -> int:
                return self.a * b

            def print(self, s: str):
                print(f'{self.a} and {s}')

        m = MyClass(a=5)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.calc(b=42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.print(s='Hi')
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.dream(b=2)

    def test_type_annotation_string(self):
        @pedantic_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        MyClass.generator()

    def test_typo_in_type_annotation_string(self):
        @pedantic_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            @staticmethod
            def generator() -> 'MyClas':
                return MyClass(s='generated')

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.generator()

    def test_overriding_contains(self):
        @pedantic_class
        class MyClass(list):
            def __contains__(self, item: int) -> bool:
                return True

        m = MyClass()
        print(42 in m)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            print('something' in m)

    def test_type_annotation_string_typo(self):
        @pedantic_class
        class MyClass:
            def compare(self, other: 'MyClas') -> bool:
                return self == other

            def fixed_compare(self, other: 'MyClass') -> bool:
                return self == other

        m = MyClass()
        m.fixed_compare(other=m)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.compare(other=m)

    def test_docstring_not_required(self):
        @pedantic_class
        class Foo:
            def __init__(self, a: int) -> None:
                self.a = a

            def bunk(self) -> int:
                '''
                Function with correct docstring. Yes, single-quoted docstrings are allowed too.
                Returns:
                    int: 42
                '''
                return self.a

        foo = Foo(a=10)
        foo.bunk()

    def test_overrides(self):
        @pedantic_class
        class Abstract:
            def func(self, b: str) -> str:
                pass

            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return 42

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()

    def test_overrides_abc(self):
        @pedantic_class
        class Abstract(ABC):
            @abstractmethod
            def func(self, b: str) -> str:
                pass

            @abstractmethod
            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return 42

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()

    def test_overrides_with_type_errors_and_call_by_args3(self):
        @pedantic_class
        class Abstract:
            def func(self, b: str) -> str:
                pass

            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return self.a

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            f.func('Hi')
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo(a=3.1415)
        f.a = 3.145
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            f.bunk()

    def test_overrides_goes_wrong(self):
        @pedantic_class
        class Parent:
            def func(self, b: str) -> str:
                return b + b + b

            def bunk(self) -> int:
                return 42

        with self.assertRaises(expected_exception=PedanticOverrideException):
            @pedantic_class
            class Foo(Parent):
                def __init__(self, a: int) -> None:
                    self.a = a

                @overrides(Parent)
                def funcy(self, b: str) -> str:
                    return b

                @overrides(Parent)
                def bunk(self) -> int:
                    return self.a

            f = Foo(a=40002)
            f.func(b='Hi')
            f.bunk()

        p = Parent()
        p.func(b='Hi')
        p.bunk()

    def test_static_method_with_sloppy_type_annotation(self):
        @pedantic_class
        class MyStaticClass:
            @staticmethod
            def double_func(a: int) -> int:
                x, y = MyStaticClass.static_bar()
                return x

            @staticmethod
            def static_bar() -> (int, int):  # this is wrong. Correct would be Tuple[int, int]
                return 0, 1

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            print(MyStaticClass.double_func(a=0))

    def test_property(self):
        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: Any) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self) -> int:
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: str) -> None:
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            MyClass(42)

        m = MyClass(some_arg=42)
        self.assertEqual(m.some_attribute, 42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_attribute = 100
        self.assertEqual(m.some_attribute, 42)
        m.some_attribute = '100'
        self.assertEqual(m._some_attribute, '100')
        m.calc(value=42.0)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            print(m.some_attribute)

    def test_property_getter_and_setter_misses_type_hints(self):
        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: int) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self):
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: int):
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            MyClass(42)

        m = MyClass(some_arg=42)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.some_attribute = 100

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            print(m.some_attribute)
        m.calc(value=42.0)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.calc(value=42)

    def test_default_constructor(self):
        @pedantic_class
        class MyClass:
            def fun(self) -> int:
                return 42
        m = MyClass()
        m.fun()

    def test_optional_callable(self):
        @pedantic_class
        class SemanticSimilarity:
            def __init__(self, post_processing: bool = True, val: Optional[Callable[[float], float]] = None) -> None:
                if post_processing is None:
                    self.post_processing = val
                else:
                    self.post_processing = lambda x: x

        SemanticSimilarity()

    def test_optional_lambda(self):
        @pedantic_class
        class SemanticSimilarity:
            def __init__(self, val: Callable[[float], float] = lambda x: x) -> None:
                self.post_processing = val

        SemanticSimilarity()

    def test_class_method_type_annotation_missing(self):
        @pedantic_class
        class MyClass:
            @classmethod
            def do(cls):
                print('i did something')

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.do()

    def test_class_method_type_annotation(self):
        @pedantic_class
        class MyClass:
            @classmethod
            def do(cls) -> None:
                print('i did something')

            @classmethod
            def calc(cls, x: Union[int, float]) -> int:
                return x * x

        MyClass.do()
        MyClass.calc(x=5)
        m = MyClass()
        m.do()

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.calc(5)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.calc(x=5.1)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            MyClass.calc('hi')

        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.calc(5)
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            m.calc(x=5.1)
        with self.assertRaises(expected_exception=PedanticCallWithArgsException):
            m.calc('hi')

    def test_dataclass_inside(self):
        """Pedantic cannot be used on dataclasses."""

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            @pedantic_class
            @dataclass
            class MyClass:
                name: str
                unit_price: float
                quantity_on_hand: int = 0

    def test_dataclass_outside(self):
        """Pedantic cannot check the constructor of dataclasses"""

        @dataclass
        @pedantic_class
        class MyClass:
            name: str
            unit_price: float
            quantity_on_hand: int = 0

            def total_cost(self) -> int:
                return self.unit_price * self.quantity_on_hand

        MyClass(name='name', unit_price=5.1)
        a = MyClass(name='name', unit_price=5.0, quantity_on_hand=42)

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            a.total_cost()

    def test_class_decorator_static_class_method(self):
        @pedantic_class
        class Foo:
            @staticmethod
            def staticmethod() -> int:
                return 'foo'

            @classmethod
            def classmethod(cls) -> int:
                return 'foo'

            def method(self) -> int:
                return 'foo'

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo.staticmethod()
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo.classmethod()
        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            Foo().method()

    def test_pedantic_class_disable_pedantic(self):
        disable_pedantic()

        @pedantic_class
        class MyClass:
            def __init__(self, pw, **kwargs):
                self._validate_str_len(new_values=kwargs)

            @staticmethod
            def _validate_str_len(new_values: Dict[str, Any]) -> None:
                return 42

            def method(pw, **kwargs):
                MyClass._validate_str_len(new_values=kwargs)

        MyClass._validate_str_len(None)
        MyClass._validate_str_len(new_values={1: 1, 2: 2})
        MyClass(name='hi', age=12, pw='123')

    def test_disable_pedantic_2(self):
        """ https://github.com/LostInDarkMath/pedantic-python-decorators/issues/37 """

        disable_pedantic()

        @pedantic_class
        class Foo:
            def __init__(self) -> None:
                self._value = 42

            def do(self) -> None:
                print(self.bar(value=self._value))

            @staticmethod
            def bar(value: int) -> int:
                return value + 75

        f = Foo()
        f.do()



import unittest

from pedantic.models.decorated_function import DecoratedFunction


class TestDecoratedFunction(unittest.TestCase):
    def test_static_method(self):
        def f_1():
            pass

        deco_f = DecoratedFunction(f_1)
        self.assertFalse(deco_f.is_static_method)

        class MyClass:
            def f_1(self):
                pass

            @staticmethod
            def f_2():
                pass

            @classmethod
            def f_3(cls):
                pass

        deco_f_1 = DecoratedFunction(MyClass.f_1)
        deco_f_2 = DecoratedFunction(MyClass.f_2)
        deco_f_3 = DecoratedFunction(MyClass.f_3)

        self.assertFalse(deco_f_1.is_static_method)
        self.assertTrue(deco_f_2.is_static_method)
        self.assertFalse(deco_f_3.is_static_method)

    def test_function_wants_args(self):
        def f_1(*args, **kwargs):
            pass

        def f_2(a, b, *args, **kwargs):
            pass

        def f_3(a, b, *args):
            pass

        def f_4(*args):
            pass

        def f_5():
            pass

        self.assertTrue(DecoratedFunction(f_1).wants_args)
        self.assertTrue(DecoratedFunction(f_2).wants_args)
        self.assertTrue(DecoratedFunction(f_3).wants_args)
        self.assertTrue(DecoratedFunction(f_4).wants_args)
        self.assertFalse(DecoratedFunction(f_5).wants_args)

        class MyClass:
            def f(self):
                pass

            @staticmethod
            def g():
                pass

        self.assertFalse(DecoratedFunction(MyClass.f).wants_args)
        self.assertFalse(DecoratedFunction(MyClass.g).wants_args)

    def test_is_property_setter(self):
        def f_1():
            pass

        self.assertFalse(DecoratedFunction(f_1).is_property_setter)

        class MyClass:
            _h = 42

            def f_1(self):
                pass

            @staticmethod
            def f_2():
                pass

        self.assertFalse(DecoratedFunction(MyClass.f_1).is_property_setter)
        self.assertFalse(DecoratedFunction(MyClass.f_2).is_property_setter)

    def test_wants_kwargs(self):
        def f_1(*args, **kwargs):
            pass

        def f_2(a, b, *args, **kwargs):
            pass

        def f_3(a, b, *args):
            pass

        def f_4(*args):
            pass

        def f_5():
            pass

        def f_6(a, b, c):
            pass

        self.assertFalse(DecoratedFunction(f_1).should_have_kwargs)
        self.assertFalse(DecoratedFunction(f_2).should_have_kwargs)
        self.assertFalse(DecoratedFunction(f_3).should_have_kwargs)
        self.assertFalse(DecoratedFunction(f_4).should_have_kwargs)
        self.assertTrue(DecoratedFunction(f_5).should_have_kwargs)
        self.assertTrue(DecoratedFunction(f_6).should_have_kwargs)

        class A:
            def f(self):
                pass

            @staticmethod
            def g():
                pass

            def __compare__(self, other):
                pass

        self.assertTrue(DecoratedFunction(A.f).should_have_kwargs)
        self.assertTrue(DecoratedFunction(A.g).should_have_kwargs)
        self.assertFalse(DecoratedFunction(A.__compare__).should_have_kwargs)

    def test_instance_method(self):
        def h():
            pass

        self.assertFalse(DecoratedFunction(h).is_instance_method)

        class A:
            def f(self):
                pass

            @staticmethod
            def g():
                pass

        self.assertTrue(DecoratedFunction(A.f).is_instance_method)
        self.assertFalse(DecoratedFunction(A.g).is_instance_method)

    def test_num_decorators(self):
        def decorator(f):
            return f

        def f_1():
            pass

        @decorator
        def f_2():
            pass

        @decorator
        @decorator
        def f_3():
            pass

        @decorator
        @decorator
        @decorator
        def f_4():
            pass

        self.assertEqual(DecoratedFunction(f_1).num_of_decorators, 0)
        self.assertEqual(DecoratedFunction(f_2).num_of_decorators, 1)
        self.assertEqual(DecoratedFunction(f_3).num_of_decorators, 2)
        self.assertEqual(DecoratedFunction(f_4).num_of_decorators, 3)



from datetime import datetime

from pedantic import overrides
from pedantic.decorators.fn_deco_validate.validators import Validator


class DatetimeIsoFormat(Validator):
    @overrides(Validator)
    def validate(self, value: str) -> datetime:
        try:
            value = datetime.fromisoformat(value)
        except (TypeError, ValueError, AttributeError):
            self.raise_exception(msg=f'invalid value: {value} is not a datetime in ISO format', value=value)

        return value



import json
from dataclasses import dataclass
from typing import List, Dict, Any
from unittest import TestCase

from flask import Flask, Response, jsonify

from pedantic import Email, overrides
from pedantic.decorators.fn_deco_validate.exceptions import ValidateException, TooManyArguments, \
    ParameterException, ExceptionDictKey, InvalidHeader
from pedantic.decorators.fn_deco_validate.fn_deco_validate import validate, ReturnAs
from pedantic.decorators.fn_deco_validate.parameters.flask_parameters import FlaskJsonParameter, FlaskFormParameter, \
    FlaskHeaderParameter, FlaskGetParameter, FlaskPathParameter, Deserializable, GenericFlaskDeserializer
from pedantic.decorators.fn_deco_validate.validators import NotEmpty, Min

JSON = 'application/json'
OK = 200
INVALID = 422
TOO_MANY_ARGS = 400


class TestFlaskParameters(TestCase):
    def test_validator_flask_json(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(
            FlaskJsonParameter(name='key', validators=[NotEmpty()]),
        )
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.route('/required')
        @validate(
            FlaskJsonParameter(name='required', required=True),
            FlaskJsonParameter(name='not_required', required=False),
            FlaskJsonParameter(name='not_required_with_default', required=False, default=42),
        )
        def required_params(required, not_required, not_required_with_default) -> Response:
            return jsonify({
                'required': required,
                'not_required': not_required,
                'not_required_with_default': not_required_with_default,
            })

        @app.route('/types')
        @validate(
            FlaskJsonParameter(name='bool_param', value_type=bool),
            FlaskJsonParameter(name='int_param', value_type=int),
            FlaskJsonParameter(name='float_param', value_type=float),
            FlaskJsonParameter(name='str_param', value_type=str),
            FlaskJsonParameter(name='list_param', value_type=list),
            FlaskJsonParameter(name='dict_param', value_type=dict),
        )
        def different_types(
                bool_param,
                int_param,
                float_param,
                str_param,
                list_param,
                dict_param,
        ) -> Response:
            return jsonify({
                'bool_param': bool_param,
                'int_param': int_param,
                'float_param': float_param,
                'str_param': str_param,
                'list_param': list_param,
                'dict_param': dict_param,
            })

        @app.route('/args')
        @validate(
            FlaskJsonParameter(name='a', validators=[NotEmpty()]),
            FlaskJsonParameter(name='b', validators=[NotEmpty()]),
            return_as=ReturnAs.ARGS,
        )
        def names_do_not_need_to_match(my_key: str, another: str) -> Response:
            return jsonify({
                'my_key': my_key,
                'another': another,
            })

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        @app.errorhandler(TooManyArguments)
        def handle_validation_error(exception: TooManyArguments) -> Response:
            print(str(exception))
            response = jsonify(str(exception))
            response.status_code = TOO_MANY_ARGS
            return response

        with app.test_client() as client:
            res = client.get('/', data=json.dumps({'key': '  hello world  '}), content_type=JSON)
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/', data=json.dumps({'key': '  '}), content_type=JSON)
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: 'NotEmpty',
                ExceptionDictKey.VALUE: '  ',
                ExceptionDictKey.MESSAGE: 'Got empty String which is invalid.',
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

            data = {
                'key': '  hello world  ',
                'required': '1',
            }
            res = client.get('/', data=json.dumps(data), content_type=JSON)
            self.assertEqual(TOO_MANY_ARGS, res.status_code)
            self.assertEqual("Got unexpected arguments: ['required']", res.json)

    def test_validator_flask_form_data(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='key', validators=[NotEmpty()]))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/', data={'key': '  hello world  '})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/', data={'key': '  '})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: 'NotEmpty',
                ExceptionDictKey.VALUE: '  ',
                ExceptionDictKey.MESSAGE: 'Got empty String which is invalid.',
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

    def test_validator_flask_header(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskHeaderParameter(name='key', validators=[NotEmpty()]))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(InvalidHeader)
        def handle_validation_error(exception: InvalidHeader) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/', headers={'key': '  hello world  '}, data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/', headers={'key': '  '}, data={})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALUE: '  ',
                ExceptionDictKey.MESSAGE: 'Got empty String which is invalid.',
                ExceptionDictKey.PARAMETER: 'key',
                ExceptionDictKey.VALIDATOR: 'NotEmpty',
            }
            self.assertEqual(expected, res.json)

    def test_validator_flask_header_optional_parameter(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskHeaderParameter(name='key', validators=[NotEmpty()], required=False))
        def hello_world(key: str = None) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get('/', headers={'key': '  hello world  '}, data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/', headers={}, data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual(None, res.json)

    def test_validator_flask_get(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskGetParameter(name='key', value_type=str, validators=[NotEmpty()]))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/?key=hello_world', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello_world', res.json)

            res = client.get('/?key=hello world', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/?key= ', data={})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: 'NotEmpty',
                ExceptionDictKey.VALUE: ' ',
                ExceptionDictKey.MESSAGE: 'Got empty String which is invalid.',
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

    def test_validator_flask_get_multiple_values_for_same_key(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskGetParameter(name='key', value_type=list, validators=[NotEmpty()]))
        def hello_world(key: List[str]) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get('/?key=hello&key=world', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual(['hello', 'world'], res.json)

    def test_validator_flask_path(self) -> None:
        app = Flask(__name__)

        @app.route('/<string:key>')
        @validate(FlaskPathParameter(name='key', validators=[NotEmpty()]))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/hello_world', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello_world', res.json)

            res = client.get('/hello world', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('hello world', res.json)

            res = client.get('/   ', data={})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: 'NotEmpty',
                ExceptionDictKey.VALUE: '   ',
                ExceptionDictKey.MESSAGE: 'Got empty String which is invalid.',
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

    def test_invalid_value_type(self) -> None:
        app = Flask(__name__)

        with self.assertRaises(expected_exception=AssertionError):
            @app.route('/')
            @validate(FlaskFormParameter(name='key', value_type=tuple))
            def hello_world(key: str) -> Response:
                return jsonify(key)

    def test_wrong_name(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='k', value_type=str))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ValidateException)
        def handle_validation_error(exception: ValidateException) -> Response:
            print(str(exception))
            response = jsonify(str(exception))
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get(data={'key': 'k'})
            self.assertEqual(INVALID, res.status_code)
            self.assertIn('Value for parameter k is required.', res.json)

    def test_default_value(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='key', value_type=str, default='42'))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get(data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('42', res.json)

    def test_not_required_allows_none_kwargs_without_none(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='key', value_type=str, required=False),
                  return_as=ReturnAs.KWARGS_WITHOUT_NONE)
        def hello_world(key: str = 'it works') -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get(data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('it works', res.json)

    def test_not_required_allows_none_kwargs_with_none(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='key', value_type=str, required=False, default=None),
                  return_as=ReturnAs.KWARGS_WITH_NONE)
        def hello_world(key: str) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get(data={'key': None})
            self.assertEqual(OK, res.status_code)
            self.assertEqual(None, res.json)

    def test_not_required_with_default(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskFormParameter(name='key', value_type=str, required=False, default='42'))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get(data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('42', res.json)

    def test_validator_flask_path_type_conversion(self) -> None:
        app = Flask(__name__)

        @app.route('/<string:key>')
        @validate(FlaskPathParameter(name='key', value_type=int, validators=[Min(42)]))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/42', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual(42, res.json)

            res = client.get('/42f', data={})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: None,
                ExceptionDictKey.VALUE: '42f',
                ExceptionDictKey.MESSAGE: "Value 42f cannot be converted to <class 'int'>.",
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

    def test_validator_flask_json_parameter_does_not_get_json(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskJsonParameter(name='key'))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            print(str(exception))
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get('/', data={})
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.VALIDATOR: None,
                ExceptionDictKey.VALUE: 'None',
                ExceptionDictKey.MESSAGE: 'Value for parameter key is required.',
                ExceptionDictKey.PARAMETER: 'key',
            }
            self.assertEqual(expected, res.json)

    def test_validator_flask_json_parameter_does_not_get_json_but_default(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskJsonParameter(name='key', default='42'))
        def hello_world(key: str) -> Response:
            return jsonify(key)

        with app.test_client() as client:
            res = client.get('/', data={})
            self.assertEqual(OK, res.status_code)
            self.assertEqual('42', res.json)

    def test_too_many_arguments(self) -> None:
        app = Flask(__name__)

        @app.route('/')
        @validate(FlaskJsonParameter(name='k', value_type=str))
        def hello_world(**kwargs) -> Response:
            return jsonify(str(kwargs))

        @app.errorhandler(TooManyArguments)
        def handle_validation_error(exception: TooManyArguments) -> Response:
            print(str(exception))
            response = jsonify(str(exception))
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get(data=json.dumps({'k': 'k', 'a': 1}), content_type=JSON)
            self.assertEqual(INVALID, res.status_code)
            self.assertEqual("Got unexpected arguments: ['a']", res.json)

    def test_exception_for_required_parameter(self) -> None:
        app = Flask(__name__)
        key = 'PASSWORD'

        @app.route('/')
        @validate(FlaskJsonParameter(name=key, value_type=str))
        def hello_world(**kwargs) -> Response:
            return jsonify(str(kwargs))

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            reason = 'required' if 'required' in exception.message else 'invalid'
            response = jsonify({exception.parameter_name: [{'KEY': reason}]})
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get(data=json.dumps({}), content_type=JSON)
            self.assertEqual(INVALID, res.status_code)
            self.assertEqual({key: [{'KEY': 'required'}]}, res.json)

    def test_async_endpoints(self) -> None:
        app = Flask(__name__)

        @app.route('/<int:k>')
        @validate(FlaskPathParameter(name='k', value_type=int, validators=[Min(42)]))
        async def hello_world(k) -> Response:
            return jsonify(str(k))

        @app.route('/foo/<int:k>')
        @validate(FlaskPathParameter(name='k', value_type=int, validators=[Min(42)]), return_as=ReturnAs.ARGS)
        async def return_args(k) -> Response:
            return jsonify(str(k))

        @app.route('/bar/<int:k>')
        @validate(FlaskPathParameter(name='k', value_type=int, validators=[Min(42)]),
                  return_as=ReturnAs.KWARGS_WITH_NONE)
        async def return_kwargs_with_none(k) -> Response:
            return jsonify(str(k))

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            res = client.get(path=f'/45')
            self.assertEqual(OK, res.status_code)

            res = client.get(path=f'/foo/45')
            self.assertEqual(OK, res.status_code)

            res = client.get(path=f'/bar/45')
            self.assertEqual(OK, res.status_code)

            res = client.get(path=f'/41')
            self.assertEqual(INVALID, res.status_code)
            expected = {
                ExceptionDictKey.MESSAGE: 'smaller then allowed: 41 is not >= 42',
                ExceptionDictKey.PARAMETER: 'k',
                ExceptionDictKey.VALIDATOR: 'Min',
                ExceptionDictKey.VALUE: '41',
            }
            self.assertEqual(expected, res.json)

    def test_json_parameter_with_default_value(self) -> None:
        app = Flask(__name__)

        @app.route('/testing/message/<string:email>', methods=['POST'])
        @validate(
            FlaskPathParameter(name='email', value_type=str, validators=[Email()]),
            FlaskJsonParameter(name='content', value_type=str, default='this is a fake message', required=False),
        )
        def testing_send_system_message(email: str, content: str) -> Response:
            return jsonify({'email': email, 'content': content})

        with app.test_client() as client:
            res = client.post(path=f'/testing/message/adam@eva.de')
            self.assertEqual(OK, res.status_code)
            expected = {
                'email': 'adam@eva.de',
                'content': 'this is a fake message',
            }
            self.assertEqual(expected, res.json)

            res = client.post(path=f'/testing/message/adam@eva.de', json={'content': 'hello world'})
            self.assertEqual(OK, res.status_code)
            expected = {
                'email': 'adam@eva.de',
                'content': 'hello world',
            }
            self.assertEqual(expected, res.json)

    def test_generic_deserializer(self) -> None:
        @dataclass(frozen=True)
        class User(Deserializable):
            firstname: str
            lastname: str
            age: int

            @staticmethod
            @overrides(Deserializable)
            def from_json(data: Dict[str, Any]) -> 'User':
                return User(
                    firstname=data['firstname'],
                    lastname=data['lastname'],
                    age=Min(value=18).validate_param(value=int(data['age']), parameter_name='age'),
                )

        app = Flask(__name__)

        @app.route('/foo', methods=['POST'])
        @validate(
            GenericFlaskDeserializer(name='user', cls=User, catch_exception=True),
        )
        def the_generic_approach(user: User) -> Response:
            return jsonify({'name': user.firstname})

        @app.route('/bar', methods=['POST'])
        @validate(
            GenericFlaskDeserializer(name='user', cls=User, catch_exception=False),
        )
        def do_not_catch(user: User) -> Response:
            return jsonify({'name': user.firstname})

        @app.errorhandler(ParameterException)
        def handle_validation_error(exception: ParameterException) -> Response:
            response = jsonify(exception.to_dict)
            response.status_code = INVALID
            return response

        with app.test_client() as client:
            data = {
                'firstname': 'Albert',
                'lastname': 'Einstein',
                'age': '56',
            }
            res = client.post(path=f'/foo', json=data)
            assert res.status_code == 200
            assert res.json == {'name': 'Albert'}

            data.pop('age')
            res = client.post(path=f'/foo', json=data)
            assert res.status_code == 422

            res = client.post(path=f'/bar', json=data)
            assert res.status_code == 500

            data['age'] = '16'
            res = client.post(path=f'/foo', json=data)
            assert res.status_code == 422
            expected = {
                ExceptionDictKey.MESSAGE: 'smaller then allowed: 16 is not >= 18',
                ExceptionDictKey.PARAMETER: 'age',
                ExceptionDictKey.VALIDATOR: 'Min',
                ExceptionDictKey.VALUE: '16',
            }
            assert res.json == expected



import sys
from typing import Type, Dict, List


def get_context(depth: int = 1, increase_depth_if_name_matches: List[str] = None) -> Dict[str, Type]:
    """
        Get the context of a frame at the given depth of the current call stack.
        See also: https://docs.python.org/3/library/sys.html#sys._getframe
    """

    frame = sys._getframe(depth)
    name = frame.f_code.co_name

    if name in (increase_depth_if_name_matches or []):
        frame = sys._getframe(depth + 1)

    return {**frame.f_globals, **frame.f_locals}



from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType
from pedantic.helper_methods import _raise_warning


def deprecated(func: F) -> F:
    """
        Use this decorator to mark a function as deprecated. It will raise a warning when the function is called.

        Example:

        >>> @deprecated
        ... def my_function(a, b, c):
        ...     pass
        >>> my_function(5, 4, 3)
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        _raise_warning(msg=f'Call to deprecated function {func.__qualname__}.', category=DeprecationWarning)
        return func(*args, **kwargs)
    return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from typing import Iterator
from unittest import TestCase

from pedantic.models import GeneratorWrapper


class TestGeneratorWrapper(TestCase):
    def test_generator_wrapper(self) -> None:
        def gen_func() -> Iterator[int]:
            num = 0

            while num < 100:
                yield num
                num += 1

        generator = gen_func()

        g = GeneratorWrapper(
            wrapped=generator,
            expected_type=Iterator[int],
            err_msg='msg',
            type_vars={},
        )

        print(sum([x for x in g]))

        with self.assertRaises(expected_exception=Exception):
            g.throw(Exception('error'))

        with self.assertRaises(expected_exception=AttributeError):
            g.invalid

        g.close()



import os

ENVIRONMENT_VARIABLE_NAME = 'ENABLE_PEDANTIC'


def enable_pedantic() -> None:
    os.environ[ENVIRONMENT_VARIABLE_NAME] = '1'


def disable_pedantic() -> None:
    os.environ[ENVIRONMENT_VARIABLE_NAME] = '0'


def is_enabled() -> bool:
    if ENVIRONMENT_VARIABLE_NAME not in os.environ:
        return True

    return os.environ[ENVIRONMENT_VARIABLE_NAME] == '1'



import unittest

from pedantic.exceptions import PedanticTypeCheckException
from pedantic.env_var_logic import enable_pedantic, disable_pedantic, is_enabled
from pedantic.decorators.fn_deco_pedantic import pedantic


class TestEnvironmentVariables(unittest.TestCase):
    def setUp(self) -> None:
        self.state = is_enabled()
        enable_pedantic()

    def tearDown(self) -> None:
        enable_pedantic()

    def test_pedantic_enabled(self):
        enable_pedantic()

        @pedantic
        def some_method():
            return 42

        with self.assertRaises(expected_exception=PedanticTypeCheckException):
            some_method()

    def test_pedantic_disabled(self):
        disable_pedantic()

        @pedantic
        def some_method():
            return 42

        some_method()

    def test_enable_disable(self):
        enable_pedantic()
        self.assertTrue(is_enabled())
        disable_pedantic()
        self.assertFalse(is_enabled())
        enable_pedantic()
        self.assertTrue(is_enabled())



from functools import wraps
from typing import Any

from pedantic.constants import F, ReturnType
from pedantic.models.decorated_function import DecoratedFunction
from pedantic.models.function_call import FunctionCall


def require_kwargs(func: F) -> F:
    """
        Checks that each passed argument is a keyword argument.

        Example:

        >>> @require_kwargs
        ... def my_function(a, b, c):
        ...     return a + b + c
        >>> my_function(5, 4, 3)
        Traceback (most recent call last):
        ...
        pedantic.exceptions.PedanticCallWithArgsException: In function my_function:
        Use kwargs when you call function my_function. Args: (5, 4, 3)
        >>> my_function(a=5, b=4, c=3)
        12
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> ReturnType:
        decorated_func = DecoratedFunction(func=func)
        call = FunctionCall(func=decorated_func, args=args, kwargs=kwargs, context={})
        call.assert_uses_kwargs()
        return func(*args, **kwargs)
    return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)
